#include "application.h"
#include "acframe_appwindow.h"
#include "acgfx_graphics.h"
#include "acutil_log.h"
#include "acgfx_font.h"
#include "acgfx_camera.h"
#include "acgfx_dynrender.h"
#include "acutil_timer.h"
#include <stdio.h>
#include <d3dx9.h>

using namespace acGraphics;

CApplication::CApplication() : CBaseApplication()
{
	font          = 0;
	camera        = 0;
	light         = 0;
	normalTexture = 0;
}

CApplication::~CApplication()
{
}

int CApplication::Init()
{
	int r;

	SetWindowText(wnd->GetHandle(), "Normal map");

	r = gfx->LoadEffect("data/font.fx");
	if( r < 0 )
	{
		LOG(("Failed to load 'data/font.fx' (%d)", r));
		return -1;
	}

	r = gfx->LoadEffect("data/scene.fx");
	if( r < 0 )
	{
		LOG(("Failed to load 'data/scene.fx' (%d)", r));
		return -1;
	}

	font = new CFont();
	r = font->Init("data/arial24.fnt", render);
	if( r < 0 )
	{
		LOG(("Failed to initialize font 'data/arial24.fnt' (%d)", r));
		return -1;
	}

	HRESULT hr = D3DXCreateTextureFromFile(gfx->GetDevice(), "data/normalmap.png", &normalTexture);
	if( FAILED(hr) )
	{
		LOG(("Failed to load texture 'data/normalmap.png' (%X)", hr));
		return -1;
	}

	camera = new CCamera(gfx);
	camera->SetPosition(D3DXVECTOR3(8, 5, 8));
	camera->SetLookAt(D3DXVECTOR3(0,0,0));
	camera->SetNearFarPlanes(0.1f, 100.0f);

	light = new CCamera(gfx);
	light->SetLookAt(D3DXVECTOR3(0,0,0));
	light->SetOrthogonal(14);
	light->SetNearFarPlanes(2.0f, 15.0f);
	light->SetAspectRatio(1);

	// Create the scene
	ID3DXMesh *mesh = 0;

	// First a small sphere to represent the light source
	D3DXCreateSphere(gfx->GetDevice(), 0.1f, 12, 8, &mesh, 0);
	meshes.push_back(mesh);
	positions.push_back(D3DXVECTOR3(3.5f, 1, .7f));

	// Then our target object
	CreateSphere();

	return 0;
}

void CApplication::Uninit()
{
	if( normalTexture )
	{
		normalTexture->Release();
		normalTexture = 0;
	}

	if( font )
	{
		delete font;
		font = 0;
	}
	
	if( light )
	{
		delete light;
		light = 0;
	}

	if( camera )
	{
		delete camera;
		camera = 0;
	}

	for( unsigned int n = 0; n < meshes.size(); n++ )
	{
		if( meshes[n] )
		{
			meshes[n]->Release();
			meshes[n] = 0;
		}
	}
}

void CApplication::Frame()
{
	float t = timer->GetFrameTime()/2;

	// Have the camera rotate the scene
	float radius = 7;
	D3DXVECTOR3 pos(sinf(t/1.22f), 
		            sinf(t*0.93f),
					cosf(t/2.2f));
	D3DXVec3Normalize(&pos, &pos);
	camera->SetPosition(radius*pos);

	// Have the light source rotate the scene
	radius = 2.5f+0.4f*sinf(t);
	pos = D3DXVECTOR3(sinf(t/1.23f), 
		              sinf(t),
					  cosf(t/2.12f));
	D3DXVec3Normalize(&pos, &pos);
	light->SetPosition(radius*pos);

	positions[0] = light->GetPosition();
}

void CApplication::RenderScene(CCamera *cam)
{
	// Render the scene
	gfx->SetEffect("data/scene.fx");

	// Set the variables needed for shadow map lookup
	gfx->SetEffectParam("LIGHTDIR", light->GetViewVector());
	gfx->SetEffectParam("LIGHTPOS", light->GetPosition());
	gfx->SetTexture(normalTexture);

	cam->SetupViewProjMtx();

	// Loop over the effect passes, and render the scene for each pass
	UINT numPasses;
	gfx->BeginEffect(&numPasses);
	for( UINT pass = 0; pass < numPasses; pass++ )
	{
		gfx->BeginEffectPass(pass);
		for( unsigned int n = 0; n < meshes.size(); n++ )
		{
			// Set the position of the mesh
			D3DXMATRIX mtx;
			D3DXMatrixTranslation(&mtx, positions[n].x, positions[n].y, positions[n].z);
			gfx->SetMatrix(D3DTS_WORLD, &mtx);
			gfx->CommitEffectChanges();

			// Render it
			meshes[n]->DrawSubset(0);
		}
		gfx->EndEffectPass();
	}
	gfx->EndEffect();
}

void CApplication::Render()
{
	IDirect3DDevice9 *dev = gfx->GetDevice();

	//------------------
	// Render the scene
	DWORD color = D3DCOLOR_ARGB(0,123,145,145);
	dev->Clear(0, 0, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, color, 1.0f, 0);
	dev->BeginScene();
	RenderScene(camera);

	//------------------------------
	// Write the FPS
	gfx->SetEffect("data/font.fx");
	char buf[256];
	sprintf(buf, "fps: %.2f", timer->GetFps());
	font->PreparePixelPerfectOutput();
	font->Write(5, 5, 0, buf, 0, 0);

	// Write the info
	int h;
	gfx->GetScreenDimensions(0, &h);
	font->Write(5, h-font->GetHeight(), 0, "Normal map test", 0, 0);

	dev->EndScene();
}

void CApplication::CreateSphere()
{
	// Define the vertex structure used for our sphere
	D3DVERTEXELEMENT9 vf[] = {
		{0,    0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
		{0,   12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,   0},
		{0,   24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
		{0,   32, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,  0},
		{0,   44, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0},
		{0xFF, 0, D3DDECLTYPE_UNUSED, 0, 0, 0}};

	struct vertex
	{
		D3DXVECTOR3 position;
		D3DXVECTOR3 normal;
		D3DXVECTOR2 texcoord;
		D3DXVECTOR3 tangent;
		D3DXVECTOR3 binormal;
	};

	// Create the mesh
	int numSlices   = 48;
	int numStacks   = 32;
	int numFaces    = numSlices*numStacks*2;
	int numVertices = (numSlices+1)*(numStacks+1);

	ID3DXMesh *mesh;
	HRESULT hr = D3DXCreateMesh(numFaces,numVertices,D3DXMESH_VB_MANAGED|D3DXMESH_IB_MANAGED,vf,gfx->GetDevice(),&mesh);
	if( SUCCEEDED(hr) )
	{
		// Lock the vertex buffer to fill it with data
		vertex *data;
		hr = mesh->LockVertexBuffer(D3DLOCK_DISCARD, (void**)&data);
		if( SUCCEEDED(hr) )
		{
			for( int stack = 0; stack <= numStacks; stack++ )
			{
				for( int slice = 0; slice <= numSlices; slice++ )
				{
					float stackPart   = 3.141592f/numStacks;
					float stackDegree = (stack-0.5f*numStacks)*stackPart;
					float slicePart   = 3.141592f*2/numSlices;
					float sliceDegree = slice*slicePart;
/*
					// Cylinder
					data->position = 2*D3DXVECTOR3(cosf(sliceDegree), 2*float(stack)/numStacks-1, sinf(sliceDegree));
					data->normal = data->position;
					data->normal.y = 0;
					D3DXVec3Normalize(&data->normal, &data->normal);
					data->texcoord = D3DXVECTOR2(float(slice)/numSlices,float(stack)/numStacks);
					data->tangent  = D3DXVECTOR3(cosf(sliceDegree+3.141592f/2),0,sinf(sliceDegree+3.141592f/2));
					D3DXVec3Cross(&data->binormal, &data->normal, &data->tangent);
					D3DXVec3Normalize(&data->binormal, &data->binormal);
*/
					// Sphere
					data->position = 2*D3DXVECTOR3(cosf(sliceDegree)*cosf(stackDegree), sinf(stackDegree), sinf(sliceDegree)*cosf(stackDegree));
					D3DXVec3Normalize(&data->normal, &data->position);
					data->texcoord = D3DXVECTOR2(float(slice)/numSlices,float(stack)/numStacks);
					data->tangent  = D3DXVECTOR3(cosf(sliceDegree+3.141592f/2),0,sinf(sliceDegree+3.141592f/2));
					D3DXVec3Cross(&data->binormal, &data->tangent, &data->normal);
					D3DXVec3Normalize(&data->binormal, &data->binormal);

					data++;
				}

				// Merge vertices
				(data-1)->position = (data-numSlices-1)->position;
			}

			mesh->UnlockVertexBuffer();
		}

		WORD *index;
		hr = mesh->LockIndexBuffer(D3DLOCK_DISCARD, (void**)&index);
		if( SUCCEEDED(hr) )
		{
			for( int stack = 0; stack < numStacks; stack++ )
			{
				for( int slice = 0; slice < numSlices; slice++ )
				{
					*index++ = stack*(numSlices+1)+slice;
					*index++ = (stack+1)*(numSlices+1)+slice;
					*index++ = stack*(numSlices+1)+slice+1;

					*index++ = stack*(numSlices+1)+slice+1;
					*index++ = (stack+1)*(numSlices+1)+slice;
					*index++ = (stack+1)*(numSlices+1)+slice+1;
				}
			}

			mesh->UnlockIndexBuffer();
		}
	}

	meshes.push_back(mesh);
	positions.push_back(D3DXVECTOR3(0,0,0));
}
