/*

    This effect will render a scene lit with a light and using a normal map

*/

shared float4x4 g_mWorld           : WORLD;           // World matrix
shared float4x4 g_mView            : VIEW;            // View matrix
shared float4x4 g_mProj            : PROJECTION;      // Projection matrix

texture  g_txNormalMap     : TEXTURE0;
float3   g_vLightDirection : LIGHTDIR;  // Used for directional light
float3   g_vLightPosition  : LIGHTPOS;  // Used for point light

sampler2D g_samNormal =
sampler_state
{
    Texture = <g_txNormalMap>;
    MinFilter = Linear;
    MagFilter = Linear;
    MipFilter = Linear;
    AddressU = Wrap;
    AddressV = Wrap;
};

struct VS_OUTPUT
{
    float4 Position   : POSITION;   // vertex position 
    float2 vTexture   : TEXCOORD0;  // texture coordinate
    float3 vNormal    : TEXCOORD1;  // normal
    float3 vTangent   : TEXCOORD2;
    float3 vBinormal  : TEXCOORD3;
    float3 vPosition  : TEXCOORD4;  // The vertex position in world space
};

VS_OUTPUT VertScene( float4 vPos         : POSITION,
                     float3 vNormal      : NORMAL,
                     float2 vTexture     : TEXCOORD0,
                     float3 vTangent     : TANGENT,
                     float3 vBinormal    : BINORMAL )
{
    VS_OUTPUT Out;

    // Determine the position in world space
    Out.vPosition = mul( vPos, g_mWorld );
    
    // Project the position onto the screen, for the pipeline
    float4 pos = mul( float4(Out.vPosition,1), g_mView );
    Out.Position = mul( pos, g_mProj );

    // Copy the texture coordinate
    Out.vTexture = vTexture;

    // Transform the normal, tangent, and binormal from object space to world space
    Out.vNormal   = normalize(mul(vNormal, (float3x3)g_mWorld));
    Out.vTangent  = normalize(mul(vTangent, (float3x3)g_mWorld));
    Out.vBinormal = normalize(mul(vBinormal, (float3x3)g_mWorld));

    return Out;
}


float4 PixScene( VS_OUTPUT In ) : COLOR0
{
    // Determine light vector 
    float3 lv = normalize(In.vPosition-g_vLightPosition);  // Point light
//  float3 lv = normalize(g_vLightDirection);              // Directional light

    // Transform light vector into tangent space
    lv = mul(float3x3(In.vTangent,In.vBinormal,In.vNormal),lv);
    
    // Compute the lighting against the normal map
    float val = max(0, -dot(lv, normalize(tex2D(g_samNormal,In.vTexture).rgb-0.5)));
    
    return float4(val,val,val,1);
}


//--------------------------------------------------------------------------------------
// Techniques
//--------------------------------------------------------------------------------------
technique RenderScene
{
    pass P0
    {
		// Turn on alpha blending
		AlphaBlendEnable = true;
		DestBlend        = INVSRCALPHA;
		SrcBlend         = SRCALPHA;
		
		// Set the vertex and pixel shaders
        VertexShader = compile vs_2_0 VertScene();
        PixelShader  = compile ps_2_0 PixScene();
    }
}
